@@ -1,10 +1,9 @@
 from __future__ import annotations
 
-from typing import Iterable, List, Dict
+from typing import Dict, Iterable, List
 
 import pandas as pd
 from pandas import DataFrame
-from pyspark.sql import SparkSession
 
 from seshat.data_class import SFrame
 from seshat.data_class.base import GroupSFrame
@@ -33,7 +32,7 @@ class DFrame(SFrame):
     def to_spf(self) -> SFrame:
         from seshat.data_class import SPFrame
 
-        spark = SparkSession.builder.appName(configs.SPARK_APP_NAME).getOrCreate()
+        spark = SPFrame.get_spark()
         return SPFrame.from_raw(spark.createDataFrame(self.data))
 
     def extend_vertically(self, other: DataFrame):
