@@ -36,6 +36,8 @@ class JobExecutionSchedule:
     until: datetime | None
     cron_expression: str | None
     run_overlap: bool = True
+    initial_run: bool = True
+    timezone: str = "UTC"
 
     def __post_init__(self):
         import re
@@ -108,8 +110,10 @@ class JobMetadata:
     requirements_type: Optional[str]
     complexity_factor: float
     data_size: float
-    secret_key: str
-    iv: str
+    code_size: float
+    secret_key: Optional[str]
+    iv: Optional[str]
+    env_vars: dict = None
 
 
 class BaseTyperCommand:
