@@ -37,6 +37,9 @@ class Saver:
     def __call__(self, sf: SFrame, *args, **kwargs):
         self.save(sf)
 
+    def calculate_complexity(self):
+        return NotImplementedError()
+
     def save(self, sf: SFrame, *args, **kwargs):
         raise NotImplementedError()
 
