@@ -1,4 +1,5 @@
 import logging
+import math
 from typing import Any, Callable, Dict, List, TypeAlias
 
 import pandas as pd
@@ -18,7 +19,17 @@ from seshat.utils.clean_json import JSONCleaner
 InputType: TypeAlias = List[Dict[str, Any]]
 OutputType: TypeAlias = List[Dict[str, Any]]
 ProcessResponseFn: TypeAlias = Callable[[InputType, InputType], OutputType]
-ProcessBatchFn: TypeAlias = Callable[[list[dict[str, Any]]], dict]
+GetExtraCtxFn: TypeAlias = Callable[[List[Dict[str, Any]]], Dict[str, Any]]
+
+
+def math_nan_to_none(row):
+    d = row.asDict()
+    for key, value in d.items():
+        if value is None:
+            continue
+        if isinstance(value, float) and math.isnan(value):
+            d[key] = None
+    return Row(**d)
 
 
 class SFrameReducer(Transformer):
@@ -39,12 +50,14 @@ class LLMInsightExtractor(SFrameReducer):
 
     Parameters
     ----------
-    get_llm_client : BaseChatModel
+    get_llm_client : () ->  BaseChatModel
         The LLM client used to generate insights.
     template_prompt : str
         The template prompt to send to the LLM. Should include placeholders for data.
     id_column : str, optional
         The column name to use as an identifier when expanding results. Required if expand_on_id is True.
+    join_cols : list[str], optional
+        The columns to use for joining the extracted insights back to the original DataFrame. If not provided, defaults to [id_column].
     template_context : str, optional
         The system context to provide to the LLM. Defaults to a basic data scientist role.
     llm_input_columns : List[str], optional
@@ -52,7 +65,7 @@ class LLMInsightExtractor(SFrameReducer):
     process_llm_json_response_fn : ProcessResponseFn, optional
         Function to process the JSON response from the LLM.
     get_extra_context : ProcessBatchFn, optional
-        Function to process batches of data before sending to the LLM. Receives the current format_args dict and the list of batch_responses, and should return a dict to update format_args.
+        Function to process data before sending to the LLM. Receives the current data and should return a dict to update format_args.
     process_llm_response : Callable, optional
         Function to process the raw LLM response before JSON parsing.
     retry : int, default=3
@@ -71,6 +84,8 @@ class LLMInsightExtractor(SFrameReducer):
         Whether to expand results based on ID column. Requires id_column.
     inject_keys : dict[str, str], optional
         Additional keys to inject into the template prompt.
+    merge_result : bool, default=True
+        Whether to merge the extracted insights back to the original DataFrame.
 
     Raises
     ------
@@ -137,10 +152,11 @@ class LLMInsightExtractor(SFrameReducer):
         get_llm_client: Callable[[], "BaseChatModel"],
         template_prompt: str,
         id_column: str = None,
+        join_cols: list[str] = None,
         template_context: str = None,
         llm_input_columns: List[str] = None,
         process_llm_json_response_fn: ProcessResponseFn = None,
-        get_extra_context: ProcessBatchFn = None,
+        get_extra_context: GetExtraCtxFn = None,
         process_llm_response: Callable = None,
         retry: int = 3,
         llm_result_cleaner: Callable = JSONCleaner().clean,
@@ -151,6 +167,7 @@ class LLMInsightExtractor(SFrameReducer):
         groupby_inject_key: str = None,
         expand_on_id: bool = False,
         inject_keys: dict[str, str] = None,
+        merge_result: bool = True,
     ):
 
         super().__init__(group_keys)
@@ -172,9 +189,11 @@ class LLMInsightExtractor(SFrameReducer):
             Your task is to analyze and provide insights about the given dataset.
         """
         )
+        self.merge_result = merge_result
 
         self.group_by_columns = group_by_columns
         self.id_column = id_column
+        self.join_cols = join_cols or [self.id_column]
         self.expand_on_id = expand_on_id
         self.static_injected_data = inject_keys
         self.process_llm_response = process_llm_response
@@ -277,11 +296,19 @@ class LLMInsightExtractor(SFrameReducer):
             prompt_kwargs = (
                 self.get_extra_context(batch_result) if self.get_extra_context else {}
             )
-            batch_result += self.extract_insight(batch_data, prompt_kwargs, **kwargs)
+            res = self.perform_extract(batch_data, prompt_kwargs, **kwargs)
+            if res:
+                batch_result += res
         return batch_result
 
+    def extract_insight_one_shot(
+        self, data: List[Dict[str, Any]], **kwargs
+    ) -> List[Dict[str, Any]]:
+        prompt_kwargs = self.get_extra_context(data) if self.get_extra_context else {}
+        return self.perform_extract(data, prompt_kwargs, **kwargs)
+
     @track
-    def extract_insight(
+    def perform_extract(
         self, data: List[Dict[str, Any]], prompt_kwargs=None, **kwargs
     ) -> List[Dict[str, Any]]:
         """
@@ -329,11 +356,17 @@ class LLMInsightExtractor(SFrameReducer):
                     mask &= default[col] == val
                 group_df = default[mask]
 
-            selected = (
-                group_df[[*self.llm_input_columns]]
-                if self.llm_input_columns
-                else group_df
-            )
+            llm_input_columns = self.llm_input_columns or group_df.columns
+            if set(llm_input_columns) - set(group_df.columns):
+                continue
+
+            selected = group_df[[*llm_input_columns]]
+
+            if (
+                self.id_column in selected.columns
+                and selected[self.id_column].isnull().all()
+            ):
+                continue
 
             # If batch mode, create chunks otherwise use whole data
             if self.batch_mode:
@@ -343,11 +376,16 @@ class LLMInsightExtractor(SFrameReducer):
                 ]
             else:
                 data = selected.to_dict("records")
-            group_name = "-".join([group.get(c) for c in self.group_by_columns])
+            group_name = (
+                "-".join([group.get(c) for c in self.group_by_columns]) if group else ""
+            )
             inputs.append({"group_name": group_name, "data": data})
 
     def reduce_df(self, default: DataFrame, **kwargs) -> Dict[str, DataFrame]:
         # Find the groups if group_by_columns set
+        if default.empty:
+            return {"default": default}
+
         groups = (
             default[[*self.group_by_columns]].drop_duplicates().to_dict("records")
             if self.group_by_columns
@@ -359,28 +397,38 @@ class LLMInsightExtractor(SFrameReducer):
         self._find_extract_inputs(default, groups, inputs)
 
         extract_func = (
-            self.extract_insight_batch if self.batch_mode else self.extract_insight
+            self.extract_insight_batch
+            if self.batch_mode
+            else self.extract_insight_one_shot
         )
         results = []
         for d in inputs:
             results.extend(extract_func(**d))
         results = pd.DataFrame(results)
 
-        if self.expand_on_id:
+        if self.expand_on_id and not results.empty:
             redundant_cols = [
                 col
                 for col in results.columns
-                if col in default.columns and col != self.id_column
+                if col in default.columns
+                and col != self.id_column
+                and col not in set(self.join_cols)
             ]
             results = (
                 results.explode(self.id_column)
                 .set_index(self.id_column)
                 .drop(columns=redundant_cols, axis=1)
-                .join(default.set_index(self.id_column), how="right")
+            )
+        if not self.merge_result:
+            return {"default": results.drop_duplicates()}
+        if not results.empty:
+            default = (
+                pd.merge(default, results, on=self.join_cols, how="left")
                 .reset_index()
+                .drop_duplicates()
             )
 
-        return {"default": results}
+        return {"default": default}
 
     def reduce_spf(
         self, default: PySparkDataFrame, **kwargs
@@ -407,6 +455,9 @@ class LLMInsightExtractor(SFrameReducer):
         # To avoid calling process group again, cache rdd and
         # use count() to trigger running process_group and cache the result.
         rdd = default.rdd.mapPartitions(process_group)
+        # Because pandas result maybe contains math nan values and
+        # these values are not valid for spark
+        rdd = rdd.map(math_nan_to_none)
         rdd.cache()
         rdd.count()
 
